from .oai.errors import (
    AuthenticationError,
    AuthenticationExpiredError,
    AuthorizationError,
    ClientError,
    ConflictError,
    ConnectionClosedError,
    ConnectionRejectedError,
    ConnectorError,
    CredentialsRevokedError,
    DefaultHandler,
    Error,
    ErrorCode,
    ErrorResponse,
    ExceptionHandler,
    HTTPHandler,
    InternalError,
    InvalidPageTokenError,
    InvalidResponseError,
    InvalidValueError,
    MissingParameterError,
    NetworkError,
    NotFoundError,
    RateLimitError,
    RequestTimeoutError,
    TransientError,
    UnknownValueError,
    UnsupportedOperationError,
    UpstreamError,
)

__all__ = [
    # SDK types (backwards-compatible)
    "Error",
    "ErrorCode",
    "ErrorResponse",
    # Base infrastructure (backwards-compatible)
    "ConnectorError",
    "DefaultHandler",
    "ExceptionHandler",
    "HTTPHandler",
    # Authentication
    "AuthenticationError",
    "AuthenticationExpiredError",
    "CredentialsRevokedError",
    # Authorization
    "AuthorizationError",
    # Transient
    "TransientError",
    "RateLimitError",
    "UpstreamError",
    "NetworkError",
    "ConnectionRejectedError",
    "ConnectionClosedError",
    "RequestTimeoutError",
    # Client
    "ClientError",
    "NotFoundError",
    "InvalidValueError",
    "UnknownValueError",
    "InvalidPageTokenError",
    "MissingParameterError",
    "ConflictError",
    # Internal
    "InternalError",
    "InvalidResponseError",
    "UnsupportedOperationError",
]
